"""A module for sugary API wrapper around the clients."""
__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

import csv
import glob
import itertools as it
import json
import os
import random
from typing import List, Union, Iterator, Any, Iterable, Dict

import numpy as np

# https://github.com/ndjson/ndjson.github.io/issues/1#issuecomment-109935996
_jsonl_ext = {'.jsonlines', '.ndjson', '.jsonl', '.jl', '.ldjson'}
_csv_ext = {'.csv', '.tcsv'}


def _sample(iterable, sampling_rate: float = None):
    for i in iterable:
        if sampling_rate is None or random.random() < sampling_rate:
            yield i


def _subsample(iterable, sampling_rate: float = None, size: int = None, **kwargs):
    yield from it.islice(_sample(iterable, sampling_rate), size)


def _input_lines(
        lines: Iterable[str] = None,
        filepath: str = None,
        read_mode: str = 'r',
        line_format: str = 'json',
        **kwargs
) -> Iterator[Union[str, bytes]]:
    """Input function that iterates over list of strings, it can be used in the Flow API.

    :param filepath: a text file that each line contains a document
    :param lines: a list of strings, each is considered as a document
    :param size: the maximum number of the documents
    :param sampling_rate: the sampling rate between [0, 1]
    :param read_mode: specifies the mode in which the file
            is opened. 'r' for reading in text mode, 'rb' for reading in binary
    :param line_format: the format of each line ``json`` or ``csv``

    .. note::
        This function should not be directly used, use :meth:`Flow.index_lines`, :meth:`Flow.search_lines` instead
    """
    if filepath:
        file_type = os.path.splitext(filepath)[1]
        with open(filepath, read_mode) as f:
            if file_type in _jsonl_ext:
                yield from _input_ndjson(f, **kwargs)
            elif file_type in _csv_ext:
                yield from _input_csv(f, **kwargs)
            else:
                yield from _subsample(f, **kwargs)
    elif lines:
        if line_format == 'json':
            yield from _input_ndjson(lines, **kwargs)
        elif line_format == 'csv':
            yield from _input_csv(lines, **kwargs)
        else:
            yield from _subsample(lines, **kwargs)
    else:
        raise ValueError('"filepath" and "lines" can not be both empty')

def _input_ndjson(
        fp: Iterable[str],
        field_resolver: Dict[str, str] = None,
        **kwargs
):
    from jina import Document

    for line in _subsample(fp, **kwargs):
        value = json.loads(line)
        if 'groundtruth' in value and 'document' in value:
            yield Document(value['document'], field_resolver), Document(value['groundtruth'], field_resolver)
        else:
            yield Document(value, field_resolver)


def _input_csv(
        fp: Iterable[str],
        field_resolver: Dict[str, str] = None,
        **kwargs
):
    from jina import Document
    lines = csv.DictReader(fp)
    for value in _subsample(lines, **kwargs):
        if 'groundtruth' in value and 'document' in value:
            yield Document(value['document'], field_resolver), Document(value['groundtruth'], field_resolver)
        else:
            yield Document(value, field_resolver)


def _input_files(
        patterns: Union[str, List[str]],
        recursive: bool = True,
        size: int = None,
        sampling_rate: float = None,
        read_mode: str = None,
) -> Iterator[Union[str, bytes]]:
    r"""Input function that iterates over files, it can be used in the Flow API.

    :param patterns: The pattern may contain simple shell-style wildcards, e.g. '\*.py', '[\*.zip, \*.gz]'
    :param recursive: If recursive is true, the pattern '**' will match any files and
                zero or more directories and subdirectories
    :param size: the maximum number of the files
    :param sampling_rate: the sampling rate between [0, 1]
    :param read_mode: specifies the mode in which the file
            is opened. 'r' for reading in text mode, 'rb' for reading in binary mode.
            If `read_mode` is None, will iterate over filenames

    .. note::
        This function should not be directly used, use :meth:`Flow.index_files`, :meth:`Flow.search_files` instead
    """
    if read_mode not in {'r', 'rb', None}:
        raise RuntimeError(f'read_mode should be "r", "rb" or None, got {read_mode}')

    def _iter_file_exts(ps):
        return it.chain.from_iterable(glob.iglob(p, recursive=recursive) for p in ps)

    d = 0
    if isinstance(patterns, str):
        patterns = [patterns]
    for g in _iter_file_exts(patterns):
        if sampling_rate is None or random.random() < sampling_rate:
            if read_mode is None:
                yield g
            elif read_mode in {'r', 'rb'}:
                with open(g, read_mode) as fp:
                    yield fp.read()
            d += 1
        if size is not None and d > size:
            break


def _input_ndarray(
        array: 'np.ndarray', axis: int = 0, size: int = None, shuffle: bool = False
) -> Iterator[Any]:
    """Input function that iterates over a numpy array, it can be used in the Flow API.

    :param array: the numpy ndarray data source
    :param axis: iterate over that axis
    :param size: the maximum number of the sub arrays
    :param shuffle: shuffle the numpy data source beforehand

    .. note::
        This function should not be directly used, use :meth:`Flow.index_ndarray`, :meth:`Flow.search_ndarray` instead
    """
    if shuffle:
        # shuffle for random query
        array = np.take(array, np.random.permutation(array.shape[0]), axis=axis)
    d = 0
    for r in array:
        yield r
        d += 1
        if size is not None and d >= size:
            break


# for back-compatibility
_input_numpy = _input_ndarray



from .base import BaseFlow
from .mixin.crud import CRUDFlowMixin


class Flow(CRUDFlowMixin, BaseFlow):
    """The synchronous version of :class:`AsyncFlow`.

    For proper usage see `this guide` <https://docs.jina.ai/chapters/flow/index.html>
    """
    pass



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from typing import Optional
import numpy as np

from jina.executors.decorators import batching, as_ndarray
from jina.executors.encoders.frameworks import BaseTFEncoder


class ImageKerasEncoder(BaseTFEncoder):
    """
    :class:`ImageKerasEncoder` encodes data from a ndarray, potentially B x (Channel x Height x Width) into a
        ndarray of `B x D`.
    Internally, :class:`ImageKerasEncoder` wraps the models from `tensorflow.keras.applications`.
    https://keras.io/applications/
    """

    def __init__(self, model_name: Optional[str] = None, img_shape: int = 96,
                 pool_strategy: str = 'avg', channel_axis: int = 1, *args, **kwargs):
        """
        :param model_name: the name of the model. Supported models include
            ``DenseNet121``, ``DenseNet169``, ``DenseNet201``,
            ``InceptionResNetV2``,
            ``InceptionV3``,
            ``MobileNet``, ``MobileNetV2``,
            ``NASNetLarge``, ``NASNetMobile``,
            ``ResNet101``, ``ResNet152``, ``ResNet50``, ``ResNet101V2``, ``ResNet152V2``, ``ResNet50V2``,
            ``VGG16``, ``VGG19``,
            ``Xception``,
        :param pool_strategy: the pooling strategy
            - `None` means that the output of the model will be the 4D tensor output of the last convolutional block.
            - `avg` means that global average pooling will be applied to the output of the last convolutional block, and
                 thus the output of the model will be a 2D tensor.
            - `max` means that global max pooling will be applied.
        :param channel_axis: the axis id of the channel, -1 indicate the color channel info at the last axis.
                If given other, then ``np.moveaxis(data, channel_axis, -1)`` is performed before :meth:`encode`.
        """
        super().__init__(*args, **kwargs)
        self.model_name = model_name or 'MobileNetV2'
        self.pool_strategy = pool_strategy
        self.img_shape = img_shape
        self.channel_axis = channel_axis

    def post_init(self):
        super().post_init()
        self.to_device()
        import tensorflow as tf
        model = getattr(tf.keras.applications, self.model_name)(
            input_shape=(self.img_shape, self.img_shape, 3),
            include_top=False,
            pooling=self.pool_strategy,
            weights='imagenet')
        model.trainable = False
        self.model = model

    @batching
    @as_ndarray
    def encode(self, data: 'np.ndarray', *args, **kwargs) -> 'np.ndarray':
        """
        :param data: a `B x (Channel x Height x Width)` numpy ``ndarray``, `B` is the size of the batch
        :return: a `B x D` numpy ``ndarray``, `D` is the output dimension
        """
        if self.channel_axis != -1:
            data = np.moveaxis(data, self.channel_axis, -1)
        return self.model(data)



import numpy as np
from jina.executors.crafters import BaseCrafter


class AudioNormalizer(BaseCrafter):
    """:class:`AudioNormalizer` normalizes the audio signal on doc-level."""

    def craft(self, blob: np.ndarray, *args, **kwargs):
        """
        Normalize signal from the audio signal.

        Reads the `ndarray` of the audio signal,
        normalizes the signal and saves the `ndarray` of the
        normalized signal in the `blob` of the Document.

        :param blob: the ndarray of the audio signal
        :param args:  Additional positional arguments
        :param kwargs: Additional keyword arguments
        :return: a Document dict with the normalized audio signal
        """
        import librosa
        signal_norm = librosa.util.normalize(blob)

        return dict(offset=0, blob=signal_norm)



from typing import Sequence, Any, Optional

from jina.executors.evaluators.rank import BaseRankingEvaluator


class FScoreEvaluator(BaseRankingEvaluator):
    """
    :class:`FScoreEvaluator` Gives the f score of a search system result. (https://en.wikipedia.org/wiki/F-score)

    :param eval_at: the point at which precision and recall are computed, if None give, will consider all the input to evaluate
    :param beta: Parameter to weight differently precision and recall. When beta is 1, the fScore corresponds to the harmonic mean
        of precision and recall
    :param args: Additional positional arguments
    :param kwargs: Additional keyword arguments
    """

    def __init__(self, eval_at: Optional[int] = None, beta: int = 1, *args, **kwargs):
        super().__init__(*args, **kwargs)
        assert beta != 0, 'fScore is not defined for beta 0'
        self.eval_at = eval_at
        self.weight = beta**2

    def evaluate(self, actual: Sequence[Any], desired: Sequence[Any], *args, **kwargs) -> float:
        """"
        :param actual: the matched document identifiers from the request as matched by jina indexers and rankers
        :param desired: the expected documents matches
        :param args: Additional positional arguments
        :param kwargs: Additional keyword arguments
        :return the evaluation metric value for the request document
        """
        if not desired or self.eval_at == 0:
            """TODO: Agree on a behavior"""
            if not desired:
                self.logger.warning('`desired` is empty: ', f'{desired}')
            if self.eval_at == 0:
                self.logger.warning('`eval_at` is invalid: ', f'{self.eval_at}')
            return 0.0

        actual_at_k = actual[:self.eval_at] if self.eval_at else actual
        common_count = len(set(actual_at_k).intersection(set(desired)))
        recall = common_count / len(desired)

        divisor = min(self.eval_at or len(desired), len(desired))

        if divisor != 0.0:
            precision = common_count / divisor
        else:
            self.logger.warning('Divided by zero, the precison is set to 0')
            precision = 0

        if precision + recall == 0:
            return 0

        return ((1 + self.weight) * precision * recall) / ((self.weight * precision) + recall)



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from . import BaseExecutableDriver


class DeleteDriver(BaseExecutableDriver):
    """Drivers inherited from this Driver will bind :meth:`delete` by default """

    def __init__(self, executor: str = None, method: str = 'delete', *args, **kwargs):
        super().__init__(executor, method, *args, **kwargs)

    def __call__(self, *args, **kwargs):
        self.exec_fn(self.req.ids)



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from abc import abstractmethod

from ..helper import cached_property


class BaseDevice:
    """:class:`BaseFrameworkExecutor` is the base class for the executors using other frameworks internally, including `tensorflow`, `pytorch`, `onnx`, `faiss` and `paddlepaddle`."""

    @cached_property
    @abstractmethod
    def device(self):
        """
        Set the device on which the executor will be running.

        ..notes:
            In the case of using GPUs, we only use the first gpu from the visible gpus. To specify which gpu to use,
            please use the environment variable `CUDA_VISIBLE_DEVICES`.
        """

    @abstractmethod
    def to_device(self, *args, **kwargs):
        """Move the computation from GPU to CPU or vice versa."""


class TorchDevice(BaseDevice):
    """
    :class:`BaseTorchDeviceHandler` implements the base class for the executors using :mod:`torch` library. The common setups go into this class.

    To implement your own executor with the :mod:`torch` library,

    .. highlight:: python
    .. code-block:: python

        class MyAwesomeTorchEncoder(BaseEncoder, BaseTorchDeviceHandler):
            def post_init(self):
                # load your awesome model
                import torchvision.models as models
                self.model = models.mobilenet_v2().features.eval()
                self.to_device(self.model)

            def encode(self, data, *args, **kwargs):
                # use your awesome model to encode/craft/score
                import torch
                torch.set_grad_enabled(False)
                
                _input = torch.as_tensor(data, device=self.device)
                _output = self.model(_input).cpu()

                return _output.numpy()

    """

    @cached_property
    def device(self):
        """
        Set the device on which the executors using :mod:`torch` library will be running.

        ..notes:
            In the case of using GPUs, we only use the first gpu from the visible gpus. To specify which gpu to use,
            please use the environment variable `CUDA_VISIBLE_DEVICES`.
        """
        import torch
        return torch.device('cuda:0') if self.on_gpu else torch.device('cpu')

    def to_device(self, model, *args, **kwargs):
        """Load the model to device."""
        model.to(self.device)


class PaddleDevice(BaseDevice):
    """
    :class:`BasePaddleExecutor` implements the base class for the executors using :mod:`paddlepaddle` library. The common setups go into this class.

    To implement your own executor with the :mod:`paddlepaddle` library,

    .. highlight:: python
    .. code-block:: python

        class MyAwesomePaddleEncoder(BasePaddleExecutor):
            def post_init(self):
                # load your awesome model
                import paddlehub as hub
                module = hub.Module(name='mobilenet_v2_imagenet')
                inputs, outputs, self.model = module.context(trainable=False)
                self.inputs_name = input_dict['image'].name
                self.outputs_name = output_dict['feature_map'].name
                self.exe = self.to_device()

            def encode(self, data, *args, **kwargs):
                # use your awesome model to encode/craft/score
                _output, *_ = self.exe.run(
                    program=self.model,
                    fetch_list=[self.outputs_name],
                    feed={self.inputs_name: data},
                    return_numpy=True
                )
                return feature_map
    """

    @cached_property
    def device(self):
        """
        Set the device on which the executors using :mod:`paddlepaddle` library will be running.

        ..notes:
            In the case of using GPUs, we only use the first gpu from the visible gpus. To specify which gpu to use,
            please use the environment variable `CUDA_VISIBLE_DEVICES`.
        """
        import paddle.fluid as fluid
        return fluid.CUDAPlace(0) if self.on_gpu else fluid.CPUPlace()

    def to_device(self):
        """Load the model to device."""
        import paddle.fluid as fluid
        return fluid.Executor(self.device)


class TFDevice(BaseDevice):
    """
    :class:`BaseTFDeviceHandler` implements the base class for the executors using :mod:`tensorflow` library. The common setups go into this class.

    To implement your own executor with the :mod:`tensorflow` library,

    .. highlight:: python
    .. code-block:: python

        class MyAwesomeTFEncoder(BaseTFDeviceHandler):
            def post_init(self):
                # load your awesome model
                self.to_device()
                import tensorflow as tf
                model = tf.keras.applications.MobileNetV2(
                    input_shape=(self.img_shape, self.img_shape, 3),
                    include_top=False,
                    pooling=self.pool_strategy,
                    weights='imagenet')
                model.trainable = False
                self.model = model

            def encode(self, data, *args, **kwargs):
                # use your awesome model to encode/craft/score
                return self.model(data)
    """

    @cached_property
    def device(self):
        """
        Set the device on which the executors using :mod:`tensorflow` library will be running.

        ..notes:
            In the case of using GPUs, we only use the first gpu from the visible gpus. To specify which gpu to use,
            please use the environment variable `CUDA_VISIBLE_DEVICES`.
        """
        import tensorflow as tf
        cpus = tf.config.experimental.list_physical_devices(device_type='CPU')
        gpus = tf.config.experimental.list_physical_devices(device_type='GPU')
        if self.on_gpu and len(gpus) > 0:
            cpus.append(gpus[0])
        return cpus

    def to_device(self):
        """Load the model to device."""
        import tensorflow as tf
        tf.config.experimental.set_visible_devices(devices=self.device)


class OnnxDevice(BaseDevice):
    """
    :class:`OnnxDevice` implements the base class for the executors using :mod:`onnxruntime` library. The common setups go into this class.

    To implement your own executor with the :mod:`onnxruntime` library,

    .. highlight:: python
    .. code-block:: python

        class MyAwesomeOnnxEncoder(BaseOnnxDeviceHandler):
            def __init__(self, output_feature, model_path, *args, **kwargs):
                super().__init__(*args, **kwargs)
                self.outputs_name = output_feature
                self.model_path = model_path

            def post_init(self):
                import onnxruntime
                self.model = onnxruntime.InferenceSession(self.model_path, None)
                self.inputs_name = self.model.get_inputs()[0].name
                self.to_device(self.model)

            def encode(self, data, *args, **kwargs):
                # use your awesome model to encode/craft/score
                results = []
                for idx in data:
                    data_encoded, *_ = self.model.run(
                        [self.outputs_name, ], {self.inputs_name: data})
                    results.append(data_encoded)
                return np.concatenate(results, axis=0)

    """

    @cached_property
    def device(self):
        """
        Set the device on which the executors using :mod:`onnxruntime` library will be running.

        ..notes:
            In the case of using GPUs, we only use the first gpu from the visible gpus. To specify which gpu to use,
            please use the environment variable `CUDA_VISIBLE_DEVICES`.
        """
        return ['CUDAExecutionProvider'] if self.on_gpu else ['CPUExecutionProvider']

    def to_device(self, model, *args, **kwargs):
        """Load the model to device."""
        model.set_providers(self.device)


class FaissDevice(BaseDevice):
    """:class:`FaissDevice` implements the base class for the executors using :mod:`faiss` library. The common setups go into this class."""

    @cached_property
    def device(self):
        """
        Set the device on which the executors using :mod:`faiss` library will be running.

        ..notes:
            In the case of using GPUs, we only use the first gpu from the visible gpus. To specify which gpu to use,
            please use the environment variable `CUDA_VISIBLE_DEVICES`.
        """
        import faiss
        # For now, consider only one GPU, do not distribute the index
        return faiss.StandardGpuResources() if self.on_gpu else None

    def to_device(self, index, *args, **kwargs):
        """Load the model to device."""
        import faiss
        device = self.device
        return faiss.index_cpu_to_gpu(device, 0, index, None) if device is not None else index


class MindsporeDevice(BaseDevice):
    """:class:`MindsporeDevice` implements the base classes for the executors using :mod:`mindspore` library. The common setups go into this class."""

    @cached_property
    def device(self):
        """
        Set the device on which the executors using :mod:`mindspore` library will be running.

        ..notes:
            In the case of using GPUs, we only use the first gpu from the visible gpus. To specify which gpu to use,
            please use the environment variable `CUDA_VISIBLE_DEVICES`.
        """
        return 'GPU' if self.on_gpu else 'CPU'

    def to_device(self):
        """Load the model to device."""
        import mindspore.context as context
        context.set_context(mode=context.GRAPH_MODE, device_target=self.device)
